import base64
import json
import sys
import unittest
from unittest import mock

import requests

from auth0.rest import RestClient, RestClientOptions

from ...exceptions import Auth0Error, RateLimitError


class TestRest(unittest.TestCase):
    def test_options_are_used_and_override(self):
        """
        This test ensures RestClientOptions are read when passed to
        RestClient's constructor by (1) configuring a timeout and (2)
        turning off Telemetry. This proves that RestClient is inheriting
        those options, and overriding it's own constructor arguments.
        """

        options = RestClientOptions(telemetry=False, timeout=0.00001, retries=10)
        rc = RestClient(jwt="a-token", telemetry=True, timeout=30, options=options)

        # Does a timeout occur as expected?
        with self.assertRaises(requests.exceptions.Timeout):
            rc.get("http://google.com")

        # Is RestClient using the RestClientOptions.timeout value properly?
        self.assertEqual(rc.options.timeout, 0.00001)

        # Is RestClient using the RestClientOptions.retries value properly?
        self.assertEqual(rc.options.retries, 10)

        # Is RestClient using the RestClientOptions.telemetry value properly?
        self.assertEqual(rc.options.telemetry, False)

        # Is RestClient using the RestClientOptions.telemetry value properly?
        self.assertEqual(
            rc.base_headers,
            {
                "Content-Type": "application/json",
                "Authorization": "Bearer a-token",
            },
        )

    def test_options_are_created_by_default(self):
        """
        This test ensures RestClientOptions are read when passed to
        RestClient's constructor by (1) configuring a timeout and (2)
        turning off Telemetry. This proves that RestClient is inheriting
        those options, and overriding it's own constructor arguments.
        """

        rc = RestClient(jwt="a-token", telemetry=False, timeout=0.00002)

        # Does a timeout occur as expected?
        with self.assertRaises(requests.exceptions.Timeout):
            rc.get("http://google.com")

        # Did RestClient create a RestClientOptions for us?
        self.assertIsNotNone(rc.options)

        # Did RestClient assign the new RestClientOptions instance the proper timeout value from the constructor?
        self.assertEqual(rc.options.timeout, 0.00002)

        # Did RestClient use the default RestClientOptions value for retries?
        self.assertEqual(rc.options.retries, 3)

        # Did RestClient assign the new RestClientOptions instance the proper telemetry value from the constructor?
        self.assertEqual(rc.options.telemetry, False)

        # Is RestClient using the RestClientOptions.telemetry value properly?
        self.assertEqual(
            rc.base_headers,
            {
                "Content-Type": "application/json",
                "Authorization": "Bearer a-token",
            },
        )

    def test_default_options_are_used(self):
        """
        This test ensures RestClientOptions are read when passed to
        RestClient's constructor by (1) configuring a timeout and (2)
        turning off Telemetry. This proves that RestClient is inheriting
        those options, and overriding it's own constructor arguments.
        """

        options = RestClientOptions()
        rc = RestClient(jwt="a-token", options=options)

        # Did RestClient store the RestClientOptions?
        self.assertIsNotNone(rc.options)

        # Did RestClientOptions use the default 5.0 timeout?
        self.assertEqual(rc.options.timeout, 5.0)

        # Did RestClientOptions use the default 3 retries?
        self.assertEqual(rc.options.retries, 3)

        # Did RestClientOptions use the default True telemetry value?
        self.assertEqual(rc.options.telemetry, True)

    def test_get_can_timeout(self):
        rc = RestClient(jwt="a-token", telemetry=False, timeout=0.00001)

        with self.assertRaises(requests.exceptions.Timeout):
            rc.get("http://google.com")

    def test_post_can_timeout(self):
        rc = RestClient(jwt="a-token", telemetry=False, timeout=0.00001)

        with self.assertRaises(requests.exceptions.Timeout):
            rc.post("http://google.com")

    def test_put_can_timeout(self):
        rc = RestClient(jwt="a-token", telemetry=False, timeout=0.00001)

        with self.assertRaises(requests.exceptions.Timeout):
            rc.put("http://google.com")

    def test_patch_can_timeout(self):
        rc = RestClient(jwt="a-token", telemetry=False, timeout=0.00001)

        with self.assertRaises(requests.exceptions.Timeout):
            rc.patch("http://google.com")

    def test_delete_can_timeout(self):
        rc = RestClient(jwt="a-token", telemetry=False, timeout=0.00001)

        with self.assertRaises(requests.exceptions.Timeout):
            rc.delete("http://google.com")

    @mock.patch("requests.get")
    def test_get_custom_timeout(self, mock_get):
        rc = RestClient(jwt="a-token", telemetry=False, timeout=(10, 2))
        headers = {
            "Authorization": "Bearer a-token",
            "Content-Type": "application/json",
        }
        mock_get.return_value.text = '["a", "b"]'
        mock_get.return_value.status_code = 200

        rc.get("the-url")
        mock_get.assert_called_with(
            "the-url", params=None, headers=headers, timeout=(10, 2)
        )

    @mock.patch("requests.post")
    def test_post_custom_timeout(self, mock_post):
        rc = RestClient(jwt="a-token", telemetry=False, timeout=(10, 2))
        headers = {
            "Authorization": "Bearer a-token",
            "Content-Type": "application/json",
        }
        mock_post.return_value.text = '["a", "b"]'
        mock_post.return_value.status_code = 200

        rc.post("the-url")
        mock_post.assert_called_with(
            "the-url", json=None, headers=headers, timeout=(10, 2)
        )

    @mock.patch("requests.put")
    def test_put_custom_timeout(self, mock_put):
        rc = RestClient(jwt="a-token", telemetry=False, timeout=(10, 2))
        headers = {
            "Authorization": "Bearer a-token",
            "Content-Type": "application/json",
        }
        mock_put.return_value.text = '["a", "b"]'
        mock_put.return_value.status_code = 200

        rc.put("the-url")
        mock_put.assert_called_with(
            "the-url", json=None, headers=headers, timeout=(10, 2)
        )

    @mock.patch("requests.patch")
    def test_patch_custom_timeout(self, mock_patch):
        rc = RestClient(jwt="a-token", telemetry=False, timeout=(10, 2))
        headers = {
            "Authorization": "Bearer a-token",
            "Content-Type": "application/json",
        }
        mock_patch.return_value.text = '["a", "b"]'
        mock_patch.return_value.status_code = 200

        rc.patch("the-url")
        mock_patch.assert_called_with(
            "the-url", json=None, headers=headers, timeout=(10, 2)
        )

    @mock.patch("requests.delete")
    def test_delete_custom_timeout(self, mock_delete):
        rc = RestClient(jwt="a-token", telemetry=False, timeout=(10, 2))
        headers = {
            "Authorization": "Bearer a-token",
            "Content-Type": "application/json",
        }
        mock_delete.return_value.text = '["a", "b"]'
        mock_delete.return_value.status_code = 200

        rc.delete("the-url")
        mock_delete.assert_called_with(
            "the-url", params={}, json=None, headers=headers, timeout=(10, 2)
        )

    @mock.patch("requests.get")
    def test_get(self, mock_get):
        rc = RestClient(jwt="a-token", telemetry=False)
        headers = {
            "Authorization": "Bearer a-token",
            "Content-Type": "application/json",
        }

        mock_get.return_value.text = '["a", "b"]'
        mock_get.return_value.status_code = 200

        response = rc.get("the-url")
        mock_get.assert_called_with(
            "the-url", params=None, headers=headers, timeout=5.0
        )

        self.assertEqual(response, ["a", "b"])

        response = rc.get(url="the/url", params={"A": "param", "B": "param"})
        mock_get.assert_called_with(
            "the/url", params={"A": "param", "B": "param"}, headers=headers, timeout=5.0
        )
        self.assertEqual(response, ["a", "b"])

        mock_get.return_value.text = ""
        response = rc.get("the/url")
        self.assertEqual(response, "")

    @mock.patch("requests.get")
    def test_get_errors(self, mock_get):
        rc = RestClient(jwt="a-token", telemetry=False)

        mock_get.return_value.text = (
            '{"statusCode": 999, "errorCode": "code", "message": "message"}'
        )
        mock_get.return_value.status_code = 999

        with self.assertRaises(Auth0Error) as context:
            rc.get("the/url")

        self.assertEqual(context.exception.status_code, 999)
        self.assertEqual(context.exception.error_code, "code")
        self.assertEqual(context.exception.message, "message")

    @mock.patch("requests.get")
    def test_get_rate_limit_error(self, mock_get):
        options = RestClientOptions(telemetry=False, retries=0)
        rc = RestClient(jwt="a-token", options=options)
        rc._skip_sleep = True

        mock_get.return_value.text = (
            '{"statusCode": 429, "errorCode": "code", "message": "message"}'
        )
        mock_get.return_value.status_code = 429
        mock_get.return_value.headers = {
            "x-ratelimit-limit": "3",
            "x-ratelimit-remaining": "6",
            "x-ratelimit-reset": "9",
        }

        with self.assertRaises(Auth0Error) as context:
            rc.get("the/url")

        self.assertEqual(context.exception.status_code, 429)
        self.assertEqual(context.exception.error_code, "code")
        self.assertEqual(context.exception.message, "message")
        self.assertIsInstance(context.exception, RateLimitError)
        self.assertEqual(context.exception.reset_at, 9)

        self.assertEqual(rc._metrics["retries"], 0)

    @mock.patch("requests.get")
    def test_get_rate_limit_error_without_headers(self, mock_get):
        options = RestClientOptions(telemetry=False, retries=1)
        rc = RestClient(jwt="a-token", options=options)

        mock_get.return_value.text = (
            '{"statusCode": 429, "errorCode": "code", "message": "message"}'
        )
        mock_get.return_value.status_code = 429

        mock_get.return_value.headers = {}
        with self.assertRaises(Auth0Error) as context:
            rc.get("the/url")

        self.assertEqual(context.exception.status_code, 429)
        self.assertEqual(context.exception.error_code, "code")
        self.assertEqual(context.exception.message, "message")
        self.assertIsInstance(context.exception, RateLimitError)
        self.assertEqual(context.exception.reset_at, -1)

        self.assertEqual(rc._metrics["retries"], 1)

    @mock.patch("requests.get")
    def test_get_rate_limit_custom_retries(self, mock_get):
        options = RestClientOptions(telemetry=False, retries=5)
        rc = RestClient(jwt="a-token", options=options)
        rc._skip_sleep = True

        mock_get.return_value.text = (
            '{"statusCode": 429, "errorCode": "code", "message": "message"}'
        )
        mock_get.return_value.status_code = 429
        mock_get.return_value.headers = {
            "x-ratelimit-limit": "3",
            "x-ratelimit-remaining": "6",
            "x-ratelimit-reset": "9",
        }

        with self.assertRaises(Auth0Error) as context:
            rc.get("the/url")

        self.assertEqual(context.exception.status_code, 429)
        self.assertEqual(context.exception.error_code, "code")
        self.assertEqual(context.exception.message, "message")
        self.assertIsInstance(context.exception, RateLimitError)
        self.assertEqual(context.exception.reset_at, 9)

        self.assertEqual(rc._metrics["retries"], 5)
        self.assertEqual(rc._metrics["retries"], len(rc._metrics["backoff"]))

    @mock.patch("requests.get")
    def test_get_rate_limit_invalid_retries_below_min(self, mock_get):
        options = RestClientOptions(telemetry=False, retries=-1)
        rc = RestClient(jwt="a-token", options=options)
        rc._skip_sleep = True

        mock_get.return_value.text = (
            '{"statusCode": 429, "errorCode": "code", "message": "message"}'
        )
        mock_get.return_value.status_code = 429
        mock_get.return_value.headers = {
            "x-ratelimit-limit": "3",
            "x-ratelimit-remaining": "6",
            "x-ratelimit-reset": "9",
        }

        with self.assertRaises(Auth0Error) as context:
            rc.get("the/url")

        self.assertEqual(context.exception.status_code, 429)
        self.assertEqual(context.exception.error_code, "code")
        self.assertEqual(context.exception.message, "message")
        self.assertIsInstance(context.exception, RateLimitError)
        self.assertEqual(context.exception.reset_at, 9)

        self.assertEqual(rc._metrics["retries"], 0)

    @mock.patch("requests.get")
    def test_get_rate_limit_invalid_retries_above_max(self, mock_get):
        options = RestClientOptions(telemetry=False, retries=11)
        rc = RestClient(jwt="a-token", options=options)
        rc._skip_sleep = True

        mock_get.return_value.text = (
            '{"statusCode": 429, "errorCode": "code", "message": "message"}'
        )
        mock_get.return_value.status_code = 429
        mock_get.return_value.headers = {
            "x-ratelimit-limit": "3",
            "x-ratelimit-remaining": "6",
            "x-ratelimit-reset": "9",
        }

        with self.assertRaises(Auth0Error) as context:
            rc.get("the/url")

        self.assertEqual(context.exception.status_code, 429)
        self.assertEqual(context.exception.error_code, "code")
        self.assertEqual(context.exception.message, "message")
        self.assertIsInstance(context.exception, RateLimitError)
        self.assertEqual(context.exception.reset_at, 9)

        self.assertEqual(rc._metrics["retries"], rc.MAX_REQUEST_RETRIES())

    @mock.patch("requests.get")
    def test_get_rate_limit_retries_use_exponential_backoff(self, mock_get):
        options = RestClientOptions(telemetry=False, retries=10)
        rc = RestClient(jwt="a-token", options=options)
        rc._skip_sleep = True

        mock_get.return_value.text = (
            '{"statusCode": 429, "errorCode": "code", "message": "message"}'
        )
        mock_get.return_value.status_code = 429
        mock_get.return_value.headers = {
            "x-ratelimit-limit": "3",
            "x-ratelimit-remaining": "6",
            "x-ratelimit-reset": "9",
        }

        with self.assertRaises(Auth0Error) as context:
            rc.get("the/url")

        self.assertEqual(context.exception.status_code, 429)
        self.assertEqual(context.exception.error_code, "code")
        self.assertEqual(context.exception.message, "message")
        self.assertIsInstance(context.exception, RateLimitError)
        self.assertEqual(context.exception.reset_at, 9)

        self.assertEqual(rc._metrics["retries"], 10)
        self.assertEqual(rc._metrics["retries"], len(rc._metrics["backoff"]))

        baseBackoff = [0]
        baseBackoffSum = 0
        finalBackoff = 0

        for i in range(0, 9):
            backoff = 100 * 2**i
            baseBackoff.append(backoff)
            baseBackoffSum += backoff

        for backoff in rc._metrics["backoff"]:
            finalBackoff += backoff

        # Assert that exponential backoff is happening.
        self.assertGreaterEqual(rc._metrics["backoff"][1], rc._metrics["backoff"][0])
        self.assertGreaterEqual(rc._metrics["backoff"][2], rc._metrics["backoff"][1])
        self.assertGreaterEqual(rc._metrics["backoff"][3], rc._metrics["backoff"][2])
        self.assertGreaterEqual(rc._metrics["backoff"][4], rc._metrics["backoff"][3])
        self.assertGreaterEqual(rc._metrics["backoff"][5], rc._metrics["backoff"][4])
        self.assertGreaterEqual(rc._metrics["backoff"][6], rc._metrics["backoff"][5])
        self.assertGreaterEqual(rc._metrics["backoff"][7], rc._metrics["backoff"][6])
        self.assertGreaterEqual(rc._metrics["backoff"][8], rc._metrics["backoff"][7])
        self.assertGreaterEqual(rc._metrics["backoff"][9], rc._metrics["backoff"][8])

        # Ensure jitter is being applied.
        self.assertNotEqual(rc._metrics["backoff"][1], baseBackoff[1])
        self.assertNotEqual(rc._metrics["backoff"][2], baseBackoff[2])
        self.assertNotEqual(rc._metrics["backoff"][3], baseBackoff[3])
        self.assertNotEqual(rc._metrics["backoff"][4], baseBackoff[4])
        self.assertNotEqual(rc._metrics["backoff"][5], baseBackoff[5])
        self.assertNotEqual(rc._metrics["backoff"][6], baseBackoff[6])
        self.assertNotEqual(rc._metrics["backoff"][7], baseBackoff[7])
        self.assertNotEqual(rc._metrics["backoff"][8], baseBackoff[8])
        self.assertNotEqual(rc._metrics["backoff"][9], baseBackoff[9])

        # Ensure subsequent delay is never less than the minimum.
        self.assertGreaterEqual(rc._metrics["backoff"][1], rc.MIN_REQUEST_RETRY_DELAY())
        self.assertGreaterEqual(rc._metrics["backoff"][2], rc.MIN_REQUEST_RETRY_DELAY())
        self.assertGreaterEqual(rc._metrics["backoff"][3], rc.MIN_REQUEST_RETRY_DELAY())
        self.assertGreaterEqual(rc._metrics["backoff"][4], rc.MIN_REQUEST_RETRY_DELAY())
        self.assertGreaterEqual(rc._metrics["backoff"][5], rc.MIN_REQUEST_RETRY_DELAY())
        self.assertGreaterEqual(rc._metrics["backoff"][6], rc.MIN_REQUEST_RETRY_DELAY())
        self.assertGreaterEqual(rc._metrics["backoff"][7], rc.MIN_REQUEST_RETRY_DELAY())
        self.assertGreaterEqual(rc._metrics["backoff"][8], rc.MIN_REQUEST_RETRY_DELAY())
        self.assertGreaterEqual(rc._metrics["backoff"][9], rc.MIN_REQUEST_RETRY_DELAY())

        # Ensure delay is never more than the maximum.
        self.assertLessEqual(rc._metrics["backoff"][0], rc.MAX_REQUEST_RETRY_DELAY())
        self.assertLessEqual(rc._metrics["backoff"][1], rc.MAX_REQUEST_RETRY_DELAY())
        self.assertLessEqual(rc._metrics["backoff"][2], rc.MAX_REQUEST_RETRY_DELAY())
        self.assertLessEqual(rc._metrics["backoff"][3], rc.MAX_REQUEST_RETRY_DELAY())
        self.assertLessEqual(rc._metrics["backoff"][4], rc.MAX_REQUEST_RETRY_DELAY())
        self.assertLessEqual(rc._metrics["backoff"][5], rc.MAX_REQUEST_RETRY_DELAY())
        self.assertLessEqual(rc._metrics["backoff"][6], rc.MAX_REQUEST_RETRY_DELAY())
        self.assertLessEqual(rc._metrics["backoff"][7], rc.MAX_REQUEST_RETRY_DELAY())
        self.assertLessEqual(rc._metrics["backoff"][8], rc.MAX_REQUEST_RETRY_DELAY())
        self.assertLessEqual(rc._metrics["backoff"][9], rc.MAX_REQUEST_RETRY_DELAY())

        # Ensure total delay sum is never more than 10s.
        self.assertLessEqual(finalBackoff, 10000)

    @mock.patch("requests.post")
    def test_post(self, mock_post):
        rc = RestClient(jwt="a-token", telemetry=False)
        headers = {
            "Authorization": "Bearer a-token",
            "Content-Type": "application/json",
        }

        mock_post.return_value.text = '{"a": "b"}'

        data = {"some": "data"}

        mock_post.return_value.status_code = 200
        response = rc.post("the/url", data=data)
        mock_post.assert_called_with("the/url", json=data, headers=headers, timeout=5.0)

        self.assertEqual(response, {"a": "b"})

    @mock.patch("requests.post")
    def test_post_errors(self, mock_post):
        rc = RestClient(jwt="a-token", telemetry=False)

        mock_post.return_value.text = (
            '{"statusCode": 999, "errorCode": "code", "message": "message"}'
        )
        mock_post.return_value.status_code = 999

        with self.assertRaises(Auth0Error) as context:
            rc.post("the-url")

        self.assertEqual(context.exception.status_code, 999)
        self.assertEqual(context.exception.error_code, "code")
        self.assertEqual(context.exception.message, "message")

    @mock.patch("requests.post")
    def test_post_errors_with_no_message_property(self, mock_post):
        rc = RestClient(jwt="a-token", telemetry=False)

        mock_post.return_value.text = json.dumps(
            {"statusCode": 999, "errorCode": "code", "error": "error"}
        )
        mock_post.return_value.status_code = 999

        with self.assertRaises(Auth0Error) as context:
            rc.post("the-url")

        self.assertEqual(context.exception.status_code, 999)
        self.assertEqual(context.exception.error_code, "code")
        self.assertEqual(context.exception.message, "error")

    @mock.patch("requests.post")
    def test_post_errors_with_no_message_or_error_property(self, mock_post):
        rc = RestClient(jwt="a-token", telemetry=False)

        mock_post.return_value.text = json.dumps(
            {"statusCode": 999, "errorCode": "code"}
        )
        mock_post.return_value.status_code = 999

        with self.assertRaises(Auth0Error) as context:
            rc.post("the-url")

        self.assertEqual(context.exception.status_code, 999)
        self.assertEqual(context.exception.error_code, "code")
        self.assertEqual(context.exception.message, "")

    @mock.patch("requests.post")
    def test_post_errors_with_message_and_error_property(self, mock_post):
        rc = RestClient(jwt="a-token", telemetry=False)

        mock_post.return_value.text = json.dumps(
            {
                "statusCode": 999,
                "errorCode": "code",
                "error": "error",
                "message": "message",
            }
        )
        mock_post.return_value.status_code = 999

        with self.assertRaises(Auth0Error) as context:
            rc.post("the-url")

        self.assertEqual(context.exception.status_code, 999)
        self.assertEqual(context.exception.error_code, "code")
        self.assertEqual(context.exception.message, "message")

    @mock.patch("requests.post")
    def test_post_error_with_code_property(self, mock_post):
        rc = RestClient(jwt="a-token", telemetry=False)

        for error_status in [400, 500, None]:
            mock_post.return_value.status_code = error_status
            mock_post.return_value.text = '{"errorCode": "e0","message": "desc"}'

            with self.assertRaises(Auth0Error) as context:
                rc.post("the-url")

            self.assertEqual(context.exception.status_code, error_status)
            self.assertEqual(context.exception.error_code, "e0")
            self.assertEqual(context.exception.message, "desc")

    @mock.patch("requests.post")
    def test_post_error_with_no_error_code(self, mock_post):
        rc = RestClient(jwt="a-token", telemetry=False)

        for error_status in [400, 500, None]:
            mock_post.return_value.status_code = error_status
            mock_post.return_value.text = '{"message": "desc"}'

            with self.assertRaises(Auth0Error) as context:
                rc.post("the-url")

            self.assertEqual(context.exception.status_code, error_status)
            self.assertEqual(context.exception.error_code, "a0.sdk.internal.unknown")
            self.assertEqual(context.exception.message, "desc")

    @mock.patch("requests.post")
    def test_post_error_with_text_response(self, mock_post):
        rc = RestClient(jwt="a-token", telemetry=False)

        for error_status in [400, 500, None]:
            mock_post.return_value.status_code = error_status
            mock_post.return_value.text = "there has been a terrible error"

            with self.assertRaises(Auth0Error) as context:
                rc.post("the-url")

            self.assertEqual(context.exception.status_code, error_status)
            self.assertEqual(context.exception.error_code, "a0.sdk.internal.unknown")
            self.assertEqual(
                context.exception.message, "there has been a terrible error"
            )

    @mock.patch("requests.post")
    def test_post_error_with_no_response_text(self, mock_post):
        rc = RestClient(jwt="a-token", telemetry=False)

        for error_status in [400, 500, None]:
            mock_post.return_value.status_code = error_status
            mock_post.return_value.text = None

            with self.assertRaises(Auth0Error) as context:
                rc.post("the-url")

            self.assertEqual(context.exception.status_code, error_status)
            self.assertEqual(context.exception.error_code, "a0.sdk.internal.unknown")
            self.assertEqual(context.exception.message, "")

    @mock.patch("requests.post")
    def test_file_post_content_type_is_none(self, mock_post):
        rc = RestClient(jwt="a-token", telemetry=False)
        headers = {"Authorization": "Bearer a-token"}
        mock_post.return_value.status_code = 200
        mock_post.return_value.text = "Success"

        data = {"some": "data"}
        files = [mock.Mock()]

        rc.file_post("the-url", data=data, files=files)

        mock_post.assert_called_once_with(
            "the-url", data=data, files=files, headers=headers, timeout=5.0
        )

    @mock.patch("requests.put")
    def test_put(self, mock_put):
        rc = RestClient(jwt="a-token", telemetry=False)
        headers = {
            "Authorization": "Bearer a-token",
            "Content-Type": "application/json",
        }

        mock_put.return_value.text = '["a", "b"]'
        mock_put.return_value.status_code = 200

        data = {"some": "data"}

        response = rc.put(url="the-url", data=data)
        mock_put.assert_called_with("the-url", json=data, headers=headers, timeout=5.0)

        self.assertEqual(response, ["a", "b"])

    @mock.patch("requests.put")
    def test_put_errors(self, mock_put):
        rc = RestClient(jwt="a-token", telemetry=False)

        mock_put.return_value.text = (
            '{"statusCode": 999, "errorCode": "code", "message": "message"}'
        )
        mock_put.return_value.status_code = 999

        with self.assertRaises(Auth0Error) as context:
            rc.put(url="the/url")

        self.assertEqual(context.exception.status_code, 999)
        self.assertEqual(context.exception.error_code, "code")
        self.assertEqual(context.exception.message, "message")

    @mock.patch("requests.patch")
    def test_patch(self, mock_patch):
        rc = RestClient(jwt="a-token", telemetry=False)
        headers = {
            "Authorization": "Bearer a-token",
            "Content-Type": "application/json",
        }

        mock_patch.return_value.text = '["a", "b"]'
        mock_patch.return_value.status_code = 200

        data = {"some": "data"}

        response = rc.patch(url="the-url", data=data)
        mock_patch.assert_called_with(
            "the-url", json=data, headers=headers, timeout=5.0
        )

        self.assertEqual(response, ["a", "b"])

    @mock.patch("requests.patch")
    def test_patch_errors(self, mock_patch):
        rc = RestClient(jwt="a-token", telemetry=False)

        mock_patch.return_value.text = (
            '{"statusCode": 999, "errorCode": "code", "message": "message"}'
        )
        mock_patch.return_value.status_code = 999

        with self.assertRaises(Auth0Error) as context:
            rc.patch(url="the/url")

        self.assertEqual(context.exception.status_code, 999)
        self.assertEqual(context.exception.error_code, "code")
        self.assertEqual(context.exception.message, "message")

    @mock.patch("requests.delete")
    def test_delete(self, mock_delete):
        rc = RestClient(jwt="a-token", telemetry=False)
        headers = {
            "Authorization": "Bearer a-token",
            "Content-Type": "application/json",
        }

        mock_delete.return_value.text = '["a", "b"]'
        mock_delete.return_value.status_code = 200

        response = rc.delete(url="the-url/ID")
        mock_delete.assert_called_with(
            "the-url/ID", headers=headers, params={}, json=None, timeout=5.0
        )

        self.assertEqual(response, ["a", "b"])

    @mock.patch("requests.delete")
    def test_delete_with_body_and_params(self, mock_delete):
        rc = RestClient(jwt="a-token", telemetry=False)
        headers = {
            "Authorization": "Bearer a-token",
            "Content-Type": "application/json",
        }

        mock_delete.return_value.text = '["a", "b"]'
        mock_delete.return_value.status_code = 200

        data = {"some": "data"}
        params = {"A": "param", "B": "param"}

        response = rc.delete(url="the-url/ID", params=params, data=data)
        mock_delete.assert_called_with(
            "the-url/ID", headers=headers, params=params, json=data, timeout=5.0
        )

        self.assertEqual(response, ["a", "b"])

    @mock.patch("requests.delete")
    def test_delete_errors(self, mock_delete):
        rc = RestClient(jwt="a-token", telemetry=False)

        mock_delete.return_value.text = (
            '{"statusCode": 999, "errorCode": "code", "message": "message"}'
        )
        mock_delete.return_value.status_code = 999

        with self.assertRaises(Auth0Error) as context:
            rc.delete(url="the-url")

        self.assertEqual(context.exception.status_code, 999)
        self.assertEqual(context.exception.error_code, "code")
        self.assertEqual(context.exception.message, "message")

    def test_disabled_telemetry(self):
        rc = RestClient(jwt="a-token", telemetry=False)
        expected_headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer a-token",
        }

        self.assertEqual(rc.base_headers, expected_headers)

    def test_enabled_telemetry(self):
        rc = RestClient(jwt="a-token", telemetry=True)

        user_agent = rc.base_headers["User-Agent"]
        auth0_client_bytes = base64.b64decode(rc.base_headers["Auth0-Client"])
        auth0_client_json = auth0_client_bytes.decode("utf-8")
        auth0_client = json.loads(auth0_client_json)
        content_type = rc.base_headers["Content-Type"]

        from auth0 import __version__ as auth0_version

        python_version = "{}.{}.{}".format(
            sys.version_info.major, sys.version_info.minor, sys.version_info.micro
        )

        client_info = {
            "name": "auth0-python",
            "version": auth0_version,
            "env": {"python": python_version},
        }

        self.assertEqual(user_agent, f"Python/{python_version}")
        self.assertEqual(auth0_client, client_info)
        self.assertEqual(content_type, "application/json")



from ..rest import RestClient


class RulesConfigs:
    """RulesConfig endpoint implementation.

    Args:
        domain (str): Your Auth0 domain, e.g: 'username.auth0.com'

        token (str): Management API v2 Token

        telemetry (bool, optional): Enable or disable Telemetry
            (defaults to True)

        timeout (float or tuple, optional): Change the requests
            connect and read timeout. Pass a tuple to specify
            both values separately or a float to set both to it.
            (defaults to 5.0 for both)

        rest_options (RestClientOptions): Pass an instance of
            RestClientOptions to configure additional RestClient
            options, such as rate-limit retries.
            (defaults to None)
    """

    def __init__(
        self,
        domain,
        token,
        telemetry=True,
        timeout=5.0,
        protocol="https",
        rest_options=None,
    ):
        self.domain = domain
        self.protocol = protocol
        self.client = RestClient(
            jwt=token, telemetry=telemetry, timeout=timeout, options=rest_options
        )

    def _url(self, id=None):
        url = f"{self.protocol}://{self.domain}/api/v2/rules-configs"
        if id is not None:
            return url + "/" + id
        return url

    def all(self):
        """Lists the config variable keys for rules.

        See: https://auth0.com/docs/api/management/v2#!/Rules_Configs/get_rules_configs
        """
        return self.client.get(self._url())

    def unset(self, key):
        """Removes the rules config for a given key.

        Args:
            key (str): rules config key to remove.

        See: https://auth0.com/docs/api/management/v2#!/Rules_Configs/delete_rules_configs_by_key
        """
        return self.client.delete(self._url(key))

    def set(self, key, value):
        """Sets the rules config for a given key.

        Args:
            key (str): rules config key to set.

            value (str): value to set for the rules config key.

        See: https://auth0.com/docs/api/management/v2#!/Rules_Configs/put_rules_configs_by_key
        """
        url = self._url(f"{key}")
        body = {"value": value}
        return self.client.put(url, data=body)



from ..rest import RestClient


class DeviceCredentials:
    """Auth0 connection endpoints

    Args:
        domain (str): Your Auth0 domain, e.g: 'username.auth0.com'

        token (str): Management API v2 Token

        telemetry (bool, optional): Enable or disable Telemetry
            (defaults to True)

        timeout (float or tuple, optional): Change the requests
            connect and read timeout. Pass a tuple to specify
            both values separately or a float to set both to it.
            (defaults to 5.0 for both)

        rest_options (RestClientOptions): Pass an instance of
            RestClientOptions to configure additional RestClient
            options, such as rate-limit retries.
            (defaults to None)
    """

    def __init__(
        self,
        domain,
        token,
        telemetry=True,
        timeout=5.0,
        protocol="https",
        rest_options=None,
    ):
        self.domain = domain
        self.protocol = protocol
        self.client = RestClient(
            jwt=token, telemetry=telemetry, timeout=timeout, options=rest_options
        )

    def _url(self, id=None):
        url = f"{self.protocol}://{self.domain}/api/v2/device-credentials"
        if id is not None:
            return f"{url}/{id}"
        return url

    def get(
        self,
        user_id,
        client_id,
        type,
        fields=None,
        include_fields=True,
        page=None,
        per_page=None,
        include_totals=False,
    ):
        """List device credentials.

        Args:
            user_id (str): The user_id of the devices to retrieve.

            client_id (str): The client_id of the devices to retrieve.

            type (str): The type of credentials (public_key, refresh_token).

            fields (list, optional): A list of fields to include or exclude
                (depending on include_fields) from the result. Leave empty to
                retrieve all fields.

            include_fields (bool, optional): True if the fields specified are
                to be included in the result, False otherwise. Defaults to True.

            page (int, optional): Page index of the results to return. First page is 0.

            per_page (int, optional): Number of results per page.

            include_totals (bool, optional): True to return results inside an object
                that contains the total result count (True) or as a direct array of
                results (False, default).

        See: https://auth0.com/docs/api/management/v2#!/Device_Credentials/get_device_credentials
        """

        params = {
            "fields": fields and ",".join(fields) or None,
            "include_fields": str(include_fields).lower(),
            "user_id": user_id,
            "client_id": client_id,
            "type": type,
            "page": page,
            "per_page": per_page,
            "include_totals": str(include_totals).lower(),
        }
        return self.client.get(self._url(), params=params)

    def create(self, body):
        """Create a device public key.

        Args:
            body (dict): parameters for creating the public key (e.g: type,
                device_name, client_id, etc).

        See: https://auth0.com/docs/api/v2#!/Device_Credentials/post_device_credentials
        """
        return self.client.post(self._url(), data=body)

    def delete(self, id):
        """Delete credential.

        Args:
            id (str):  The id of the credential to delete.

        See: https://auth0.com/docs/api/management/v2#!/Device_Credentials/delete_device_credentials_by_id
        """
        return self.client.delete(self._url(id))



import aiohttp

from auth0.rest_async import AsyncRestClient


def _gen_async(client, method):
    m = getattr(client, method)

    async def closure(*args, **kwargs):
        return await m(*args, **kwargs)

    return closure


def asyncify(cls):
    methods = [
        func
        for func in dir(cls)
        if callable(getattr(cls, func)) and not func.startswith("_")
    ]

    class AsyncClient(cls):
        def __init__(
            self,
            domain,
            token,
            telemetry=True,
            timeout=5.0,
            protocol="https",
            rest_options=None,
        ):
            if token is None:
                # Wrap the auth client
                super().__init__(domain, telemetry, timeout, protocol)
            else:
                # Wrap the mngtmt client
                super().__init__(
                    domain, token, telemetry, timeout, protocol, rest_options
                )
            self.client = AsyncRestClient(
                jwt=token, telemetry=telemetry, timeout=timeout, options=rest_options
            )

    class Wrapper(cls):
        def __init__(
            self,
            domain,
            token=None,
            telemetry=True,
            timeout=5.0,
            protocol="https",
            rest_options=None,
        ):
            if token is None:
                # Wrap the auth client
                super().__init__(domain, telemetry, timeout, protocol)
            else:
                # Wrap the mngtmt client
                super().__init__(
                    domain, token, telemetry, timeout, protocol, rest_options
                )

            self._async_client = AsyncClient(
                domain, token, telemetry, timeout, protocol, rest_options
            )
            for method in methods:
                setattr(
                    self,
                    f"{method}_async",
                    _gen_async(self._async_client, method),
                )

        def set_session(self, session):
            """Set Client Session to improve performance by reusing session.

            Args:
                session (aiohttp.ClientSession): The client session which should be closed
                    manually or within context manager.
            """
            self._session = session
            self._async_client.client.set_session(self._session)

        async def __aenter__(self):
            """Automatically create and set session within context manager."""
            self.set_session(aiohttp.ClientSession())
            return self

        async def __aexit__(self, exc_type, exc_val, exc_tb):
            """Automatically close session within context manager."""
            await self._session.close()

    return Wrapper



from ..rest import RestClient


class Emails:
    """Auth0 email endpoints

    Args:
        domain (str): Your Auth0 domain, e.g: 'username.auth0.com'

        token (str): Management API v2 Token

        telemetry (bool, optional): Enable or disable Telemetry
            (defaults to True)

        timeout (float or tuple, optional): Change the requests
            connect and read timeout. Pass a tuple to specify
            both values separately or a float to set both to it.
            (defaults to 5.0 for both)

        rest_options (RestClientOptions): Pass an instance of
            RestClientOptions to configure additional RestClient
            options, such as rate-limit retries.
            (defaults to None)
    """

    def __init__(
        self,
        domain,
        token,
        telemetry=True,
        timeout=5.0,
        protocol="https",
        rest_options=None,
    ):
        self.domain = domain
        self.protocol = protocol
        self.client = RestClient(
            jwt=token, telemetry=telemetry, timeout=timeout, options=rest_options
        )

    def _url(self, id=None):
        url = f"{self.protocol}://{self.domain}/api/v2/emails/provider"
        if id is not None:
            return f"{url}/{id}"
        return url

    def get(self, fields=None, include_fields=True):
        """Get the email provider.

        Args:
            fields (list of str, optional): A list of fields to include or
                exclude from the result (depending on include_fields). Leave empty to
                retrieve all fields.

            include_fields (bool, optional): True if the fields specified are
                to be included in the result, False otherwise. Defaults to True.

        See: https://auth0.com/docs/api/management/v2#!/Emails/get_provider
        """
        params = {
            "fields": fields and ",".join(fields) or None,
            "include_fields": str(include_fields).lower(),
        }

        return self.client.get(self._url(), params=params)

    def config(self, body):
        """Configure the email provider.

        Args:
            body (dict): attributes of the created email provider.

        See: https://auth0.com/docs/api/v2#!/Emails/post_provider
        """
        return self.client.post(self._url(), data=body)

    def delete(self):
        """Delete the email provider. (USE WITH CAUTION)

        See: https://auth0.com/docs/api/management/v2#!/Emails/delete_provider
        """
        return self.client.delete(self._url())

    def update(self, body):
        """Update the email provider.

        Args:
            body (dict): attributes to update on the email provider

        See: https://auth0.com/docs/api/v2#!/Emails/patch_provider
        """
        return self.client.patch(self._url(), data=body)



from .base import AuthenticationBase
from .client_authentication import add_client_authentication


class GetToken(AuthenticationBase):

    """/oauth/token related endpoints

    Args:
        domain (str): Your auth0 domain (e.g: username.auth0.com)
    """

    def authorization_code(
        self,
        code,
        redirect_uri,
        grant_type="authorization_code",
    ):
        """Authorization code grant

        This is the OAuth 2.0 grant that regular web apps utilize in order
        to access an API. Use this endpoint to exchange an Authorization Code
        for a Token.

        Args:
            code (str): The Authorization Code received from the /authorize Calls

            redirect_uri (str, optional): This is required only if it was set at
            the GET /authorize endpoint. The values must match

            grant_type (str): Denotes the flow you're using. For authorization code
            use authorization_code

        Returns:
            access_token, id_token
        """

        return self.authenticated_post(
            f"{self.protocol}://{self.domain}/oauth/token",
            data={
                "client_id": self.client_id,
                "code": code,
                "grant_type": grant_type,
                "redirect_uri": redirect_uri,
            },
        )

    def authorization_code_pkce(
        self,
        code_verifier,
        code,
        redirect_uri,
        grant_type="authorization_code",
    ):
        """Authorization code pkce grant

        This is the OAuth 2.0 grant that mobile apps utilize in order to access an API.
        Use this endpoint to exchange an Authorization Code for a Token.

        Args:
            code_verifier (str): Cryptographically random key that was used to generate
            the code_challenge passed to /authorize.

            code (str): The Authorization Code received from the /authorize Calls

            redirect_uri (str, optional): This is required only if it was set at
            the GET /authorize endpoint. The values must match

            grant_type (str): Denotes the flow you're using. For authorization code pkce
            use authorization_code

        Returns:
            access_token, id_token
        """

        return self.post(
            f"{self.protocol}://{self.domain}/oauth/token",
            data={
                "client_id": self.client_id,
                "code_verifier": code_verifier,
                "code": code,
                "grant_type": grant_type,
                "redirect_uri": redirect_uri,
            },
        )

    def client_credentials(
        self,
        audience,
        grant_type="client_credentials",
    ):
        """Client credentials grant

        This is the OAuth 2.0 grant that server processes utilize in
        order to access an API. Use this endpoint to directly request
        an access_token by using the Application Credentials (a Client Id and
        a Client Secret).

        Args:
            audience (str): The unique identifier of the target API you want to access.

            grant_type (str, optional): Denotes the flow you're using. For client credentials use "client_credentials"

        Returns:
            access_token
        """

        return self.authenticated_post(
            f"{self.protocol}://{self.domain}/oauth/token",
            data={
                "client_id": self.client_id,
                "audience": audience,
                "grant_type": grant_type,
            },
        )

    def login(
        self,
        username,
        password,
        scope,
        realm,
        audience,
        grant_type="http://auth0.com/oauth/grant-type/password-realm",
    ):
        """Calls /oauth/token endpoint with password-realm grant type


        This is the OAuth 2.0 grant that highly trusted apps utilize in order
        to access an API. In this flow the end-user is asked to fill in credentials
        (username/password) typically using an interactive form in the user-agent
        (browser). This information is later on sent to the client and Auth0.
        It is therefore imperative that the client is absolutely trusted with
        this information.

        Args:
            audience (str): The unique identifier of the target API you want to access.

            username (str): Resource owner's identifier

            password (str): resource owner's Secret

            scope(str): String value of the different scopes the client is asking for.
            Multiple scopes are separated with whitespace.

            realm (str): String value of the realm the user belongs.
            Set this if you want to add realm support at this grant.

            grant_type (str, optional): Denotes the flow you're using. For password realm
            use http://auth0.com/oauth/grant-type/password-realm

        Returns:
            access_token, id_token
        """

        return self.authenticated_post(
            f"{self.protocol}://{self.domain}/oauth/token",
            data={
                "client_id": self.client_id,
                "username": username,
                "password": password,
                "realm": realm,
                "scope": scope,
                "audience": audience,
                "grant_type": grant_type,
            },
        )

    def refresh_token(
        self,
        refresh_token,
        scope="",
        grant_type="refresh_token",
    ):
        """Calls /oauth/token endpoint with refresh token grant type

        Use this endpoint to refresh an access token, using the refresh token you got during authorization.

        Args:
            refresh_token (str): The refresh token returned from the initial token request.

            scope (str): Use this to limit the scopes of the new access token.
            Multiple scopes are separated with whitespace.

            grant_type (str): Denotes the flow you're using. For refresh token
            use refresh_token

        Returns:
            access_token, id_token
        """

        return self.authenticated_post(
            f"{self.protocol}://{self.domain}/oauth/token",
            data={
                "client_id": self.client_id,
                "refresh_token": refresh_token,
                "scope": scope,
                "grant_type": grant_type,
            },
        )

    def passwordless_login(self, username, otp, realm, scope, audience):
        """Calls /oauth/token endpoint with http://auth0.com/oauth/grant-type/passwordless/otp grant type

        Once the verification code was received, login the user using this endpoint with their
        phone number/email and verification code.

        Args:
            username (str): The user's phone number or email address.

            otp (str): the user's verification code.

            realm (str): use 'sms' or 'email'.
            Should be the same as the one used to start the passwordless flow.

            scope(str): String value of the different scopes the client is asking for.
            Multiple scopes are separated with whitespace.

            audience (str): The unique identifier of the target API you want to access.

        Returns:
            access_token, id_token
        """

        return self.authenticated_post(
            f"{self.protocol}://{self.domain}/oauth/token",
            data={
                "client_id": self.client_id,
                "username": username,
                "otp": otp,
                "realm": realm,
                "scope": scope,
                "audience": audience,
                "grant_type": "http://auth0.com/oauth/grant-type/passwordless/otp",
            },
        )



import unittest
from unittest import mock

from ...management.email_templates import EmailTemplates


class TestClients(unittest.TestCase):
    def test_init_with_optionals(self):
        t = EmailTemplates(
            domain="domain", token="jwttoken", telemetry=False, timeout=(10, 2)
        )
        self.assertEqual(t.client.options.timeout, (10, 2))
        telemetry_header = t.client.base_headers.get("Auth0-Client", None)
        self.assertEqual(telemetry_header, None)

    @mock.patch("auth0.management.email_templates.RestClient")
    def test_create(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = EmailTemplates(domain="domain", token="jwttoken")
        c.create({"a": "b", "c": "d"})

        mock_instance.post.assert_called_with(
            "https://domain/api/v2/email-templates", data={"a": "b", "c": "d"}
        )

    @mock.patch("auth0.management.email_templates.RestClient")
    def test_get(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = EmailTemplates(domain="domain", token="jwttoken")
        c.get("this-template-name")

        mock_instance.get.assert_called_with(
            "https://domain/api/v2/email-templates/this-template-name"
        )

    @mock.patch("auth0.management.email_templates.RestClient")
    def test_update(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = EmailTemplates(domain="domain", token="jwttoken")
        c.update("this-template-name", {"a": "b", "c": "d"})

        mock_instance.patch.assert_called_with(
            "https://domain/api/v2/email-templates/this-template-name",
            data={"a": "b", "c": "d"},
        )



from ..rest import RestClient


class Actions:
    """Auth0 Actions endpoints

    Args:
        domain (str): Your Auth0 domain, e.g: 'username.auth0.com'

        token (str): Management API v2 Token

        telemetry (bool, optional): Enable or disable Telemetry
            (defaults to True)

        timeout (float or tuple, optional): Change the requests
            connect and read timeout. Pass a tuple to specify
            both values separately or a float to set both to it.
            (defaults to 5.0 for both)

        rest_options (RestClientOptions): Pass an instance of
            RestClientOptions to configure additional RestClient
            options, such as rate-limit retries.
            (defaults to None)
    """

    def __init__(
        self,
        domain,
        token,
        telemetry=True,
        timeout=5.0,
        protocol="https",
        rest_options=None,
    ):
        self.domain = domain
        self.protocol = protocol
        self.client = RestClient(
            jwt=token, telemetry=telemetry, timeout=timeout, options=rest_options
        )

    def _url(self, *args):
        url = f"{self.protocol}://{self.domain}/api/v2/actions"
        for p in args:
            if p is not None:
                url = f"{url}/{p}"
        return url

    def get_actions(
        self,
        trigger_id=None,
        action_name=None,
        deployed=None,
        installed=False,
        page=None,
        per_page=None,
    ):
        """Get all actions.

        Args:
           trigger_id (str, optional): Filter the results to only actions associated
                 with this trigger ID.

           action_name (str, optional): Filter the results to only actions with this name.

           deployed (bool, optional): True to filter the results to only deployed actions.
                Defaults to False.

           installed (bool, optional): True to filter the results to only installed actions.
                Defaults to False.

           page (int, optional): The result's page number (zero based). When not set,
                the default value is up to the server.

           per_page (int, optional): The amount of entries per page. When not set,
                the default value is up to the server.

        See: https://auth0.com/docs/api/management/v2#!/Actions/get_actions
        """

        if deployed is not None:
            deployed = str(deployed).lower()

        params = {
            "triggerId": trigger_id,
            "actionName": action_name,
            "deployed": deployed,
            "installed": str(installed).lower(),
            "page": page,
            "per_page": per_page,
        }

        return self.client.get(self._url("actions"), params=params)

    def create_action(self, body):
        """Create a new action.

        Args:
           body (dict): Attributes for the new action.

        See: https://auth0.com/docs/api/management/v2#!/Actions/post_action
        """

        return self.client.post(self._url("actions"), data=body)

    def update_action(self, id, body):
        """Updates an action.

        Args:
           id (str): the ID of the action.

           body (dict): Attributes to modify.

        See: https://auth0.com/docs/api/management/v2#!/Actions/patch_action
        """

        return self.client.patch(self._url("actions", id), data=body)

    def get_action(self, id):
        """Retrieves an action by its ID.

        Args:
           id (str): Id of action to retrieve.

        See: https://auth0.com/docs/api/management/v2#!/Actions/get_action
        """
        params = {}

        return self.client.get(self._url("actions", id), params=params)

    def delete_action(self, id, force=False):
        """Deletes an action and all of its associated versions.

        Args:
           id (str): ID of the action to delete.

           force (bool, optional): True to force action deletion detaching bindings,
               False otherwise. Defaults to False.

        See: https://auth0.com/docs/api/management/v2#!/Actions/delete_action
        """
        params = {"force": str(force).lower()}

        return self.client.delete(self._url("actions", id), params=params)

    def get_triggers(self):
        """Retrieve the set of triggers currently available within actions.

        See: https://auth0.com/docs/api/management/v2#!/Actions/get_triggers
        """
        params = {}

        return self.client.get(self._url("triggers"), params=params)

    def get_execution(self, id):
        """Get information about a specific execution of a trigger.

        Args:
           id (str): The ID of the execution to retrieve.

        See: https://auth0.com/docs/api/management/v2#!/Actions/get_execution
        """
        params = {}

        return self.client.get(self._url("executions", id), params=params)

    def get_action_versions(self, id, page=None, per_page=None):
        """Get all of an action's versions.

        Args:
           id (str): The ID of the action.

           page (int, optional): The result's page number (zero based). When not set,
                the default value is up to the server.

           per_page (int, optional): The amount of entries per page. When not set,
                the default value is up to the server.

        See: https://auth0.com/docs/api/management/v2#!/Actions/get_action_versions
        """
        params = {"page": page, "per_page": per_page}

        return self.client.get(self._url("actions", id, "versions"), params=params)

    def get_trigger_bindings(self, id, page=None, per_page=None):
        """Get the actions that are bound to a trigger.

        Args:
           id (str): The trigger ID.

           page (int, optional): The result's page number (zero based). When not set,
                the default value is up to the server.

           per_page (int, optional): The amount of entries per page. When not set,
                the default value is up to the server.

        See: https://auth0.com/docs/api/management/v2#!/Actions/get_bindings
        """
        params = {"page": page, "per_page": per_page}
        return self.client.get(self._url("triggers", id, "bindings"), params=params)

    def get_action_version(self, action_id, version_id):
        """Retrieve a specific version of an action.

        Args:
           action_id (str): The ID of the action.

           version_id (str): The ID of the version to retrieve.

        See: https://auth0.com/docs/api/management/v2#!/Actions/get_action_version
        """
        params = {}

        return self.client.get(
            self._url("actions", action_id, "versions", version_id), params=params
        )

    def deploy_action(self, id):
        """Deploy an action.

        Args:
           id (str): The ID of the action to deploy.

        See: https://auth0.com/docs/api/management/v2#!/Actions/post_deploy_action
        """
        return self.client.post(self._url("actions", id, "deploy"))

    def rollback_action_version(self, action_id, version_id):
        """Roll back to a previous version of an action.

        Args:
           action_id (str): The ID of the action.

           version_id (str): The ID of the version.

        See: https://auth0.com/docs/api/management/v2#!/Actions/post_deploy_draft_version
        """
        return self.client.post(
            self._url("actions", action_id, "versions", version_id, "deploy"), data={}
        )

    def update_trigger_bindings(self, id, body):
        """Update a trigger's bindings.

        Args:
           id (str): The ID of the trigger to update.

           body (dict): Attributes for the updated trigger binding.

        See: https://auth0.com/docs/api/management/v2#!/Actions/patch_bindings
        """
        return self.client.patch(self._url("triggers", id, "bindings"), data=body)



import warnings

from ..rest import RestClient


class Jobs:
    """Auth0 jobs endpoints

    Args:
        domain (str): Your Auth0 domain, e.g: 'username.auth0.com'

        token (str): Management API v2 Token

        telemetry (bool, optional): Enable or disable Telemetry
            (defaults to True)

        timeout (float or tuple, optional): Change the requests
            connect and read timeout. Pass a tuple to specify
            both values separately or a float to set both to it.
            (defaults to 5.0 for both)

        rest_options (RestClientOptions): Pass an instance of
            RestClientOptions to configure additional RestClient
            options, such as rate-limit retries.
            (defaults to None)
    """

    def __init__(
        self,
        domain,
        token,
        telemetry=True,
        timeout=5.0,
        protocol="https",
        rest_options=None,
    ):
        self.domain = domain
        self.protocol = protocol
        self.client = RestClient(
            jwt=token, telemetry=telemetry, timeout=timeout, options=rest_options
        )

    def _url(self, path=None):
        url = f"{self.protocol}://{self.domain}/api/v2/jobs"
        if path is not None:
            return f"{url}/{path}"
        return url

    def get(self, id):
        """Retrieves a job. Useful to check its status.

        Args:
            id (str): The id of the job.

        See: https://auth0.com/docs/api/management/v2#!/Jobs/get_jobs_by_id
        """
        return self.client.get(self._url(id))

    def get_failed_job(self, id):
        """Get failed job error details.

        Args:
            id (str): The id of the job.

        See: https://auth0.com/docs/api/management/v2#!/Jobs/get_errors
        """
        url = self._url(f"{id}/errors")
        return self.client.get(url)

    def export_users(self, body):
        """Export all users to a file using a long running job.

        Check job status with get(). URL pointing to the export file will be
        included in the status once the job is complete.

        Args:
            body (dict): The details of the export users request.

        See: https://auth0.com/docs/api/management/v2#!/Jobs/post_users_exports
        """
        return self.client.post(self._url("users-exports"), data=body)

    def import_users(
        self,
        connection_id,
        file_obj,
        upsert=False,
        send_completion_email=True,
        external_id=None,
    ):
        """Imports users to a connection from a file.

        Args:
            connection_id (str): The connection id of the connection to which
                users will be inserted.

            file_obj (file): A file-like object to upload. The format for
                this file is explained in: https://auth0.com/docs/bulk-import.

            upsert (bool, optional): When set to False, pre-existing users that match on email address, user ID, or username
                will fail. When set to True, pre-existing users that match on any of these fields will be updated, but
                only with upsertable attributes. Defaults to False.
                For a list of user profile fields that can be upserted during import, see the following article
                https://auth0.com/docs/users/references/user-profile-structure#user-profile-attributes.

            send_completion_email (bool, optional): When set to True, an email will be sent to notify the completion of this job.
                When set to False, no email will be sent. Defaults to True.

            external_id (str, optional):  Customer-defined ID.

        See: https://auth0.com/docs/api/management/v2#!/Jobs/post_users_imports
        """
        return self.client.file_post(
            self._url("users-imports"),
            data={
                "connection_id": connection_id,
                "upsert": str(upsert).lower(),
                "send_completion_email": str(send_completion_email).lower(),
                "external_id": external_id,
            },
            files={"users": file_obj},
        )

    def send_verification_email(self, body):
        """Send verification email.

        Send an email to the specified user that asks them to click a link to
        verify their email address.

        Args:
            body (dict): Details of verification email request.

        See: https://auth0.com/docs/api/v2#!/Jobs/post_verification_email
        """
        return self.client.post(self._url("verification-email"), data=body)



import unittest
from unittest import mock

from ...management.jobs import Jobs


class TestJobs(unittest.TestCase):
    def test_init_with_optionals(self):
        t = Jobs(domain="domain", token="jwttoken", telemetry=False, timeout=(10, 2))
        self.assertEqual(t.client.options.timeout, (10, 2))
        telemetry_header = t.client.base_headers.get("Auth0-Client", None)
        self.assertEqual(telemetry_header, None)

    @mock.patch("auth0.management.jobs.RestClient")
    def test_get(self, mock_rc):
        mock_instance = mock_rc.return_value

        j = Jobs(domain="domain", token="jwttoken")
        j.get("an-id")

        mock_instance.get.assert_called_with(
            "https://domain/api/v2/jobs/an-id",
        )

    @mock.patch("auth0.management.jobs.RestClient")
    def test_get_failed_job(self, mock_rc):
        mock_instance = mock_rc.return_value

        j = Jobs(domain="domain", token="jwttoken")
        j.get_failed_job("an-id")

        mock_instance.get.assert_called_with(
            "https://domain/api/v2/jobs/an-id/errors",
        )

    @mock.patch("auth0.management.jobs.RestClient")
    def test_export_users(self, mock_rc):
        mock_instance = mock_rc.return_value

        j = Jobs(domain="domain", token="jwttoken")
        j.export_users({"connection_id": "cxn_id", "format": "json"})

        mock_instance.post.assert_called_with(
            "https://domain/api/v2/jobs/users-exports",
            data={"connection_id": "cxn_id", "format": "json"},
        )

    @mock.patch("auth0.management.jobs.RestClient")
    def test_import_users(self, mock_rc):
        mock_instance = mock_rc.return_value

        j = Jobs(domain="domain", token="jwttoken")
        j.import_users(connection_id="1234", file_obj={})

        mock_instance.file_post.assert_called_with(
            "https://domain/api/v2/jobs/users-imports",
            data={
                "connection_id": "1234",
                "upsert": "false",
                "send_completion_email": "true",
                "external_id": None,
            },
            files={"users": {}},
        )

        j.import_users(
            connection_id="1234",
            file_obj={},
            upsert=True,
            send_completion_email=False,
            external_id="ext-id-123",
        )
        mock_instance.file_post.assert_called_with(
            "https://domain/api/v2/jobs/users-imports",
            data={
                "connection_id": "1234",
                "upsert": "true",
                "send_completion_email": "false",
                "external_id": "ext-id-123",
            },
            files={"users": {}},
        )

        j.import_users(
            connection_id="1234", file_obj={}, upsert=False, send_completion_email=True
        )
        mock_instance.file_post.assert_called_with(
            "https://domain/api/v2/jobs/users-imports",
            data={
                "connection_id": "1234",
                "upsert": "false",
                "send_completion_email": "true",
                "external_id": None,
            },
            files={"users": {}},
        )

    @mock.patch("auth0.management.jobs.RestClient")
    def test_verification_email(self, mock_rc):
        mock_instance = mock_rc.return_value

        j = Jobs(domain="domain", token="jwttoken")
        j.send_verification_email({"a": "b", "c": "d"})

        mock_instance.post.assert_called_with(
            "https://domain/api/v2/jobs/verification-email", data={"a": "b", "c": "d"}
        )



import unittest
from unittest import mock

from ...management.rules_configs import RulesConfigs


class TestRulesConfigs(unittest.TestCase):
    def test_init_with_optionals(self):
        t = RulesConfigs(
            domain="domain", token="jwttoken", telemetry=False, timeout=(10, 2)
        )
        self.assertEqual(t.client.options.timeout, (10, 2))
        telemetry_header = t.client.base_headers.get("Auth0-Client", None)
        self.assertEqual(telemetry_header, None)

    @mock.patch("auth0.management.rules_configs.RestClient")
    def test_all(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = RulesConfigs(domain="domain", token="jwttoken")

        c.all()

        args, kwargs = mock_instance.get.call_args

        self.assertEqual("https://domain/api/v2/rules-configs", args[0])

    @mock.patch("auth0.management.rules_configs.RestClient")
    def test_unset(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = RulesConfigs(domain="domain", token="jwttoken")
        c.unset("an-id")

        mock_instance.delete.assert_called_with(
            "https://domain/api/v2/rules-configs/an-id"
        )

    @mock.patch("auth0.management.rules_configs.RestClient")
    def test_set(self, mock_rc):
        mock_instance = mock_rc.return_value

        g = RulesConfigs(domain="domain", token="jwttoken")
        g.set("key", "MY_RULES_CONFIG_VALUES")

        args, kwargs = mock_instance.put.call_args
        self.assertEqual("https://domain/api/v2/rules-configs/key", args[0])



import unittest
from unittest import mock

from ...management.tickets import Tickets


class TestTickets(unittest.TestCase):
    def test_init_with_optionals(self):
        t = Tickets(domain="domain", token="jwttoken", telemetry=False, timeout=(10, 2))
        self.assertEqual(t.client.options.timeout, (10, 2))
        telemetry_header = t.client.base_headers.get("Auth0-Client", None)
        self.assertEqual(telemetry_header, None)

    @mock.patch("auth0.management.tickets.RestClient")
    def test_email(self, mock_rc):
        mock_instance = mock_rc.return_value

        t = Tickets(domain="domain", token="jwttoken")
        t.create_email_verification({"a": "b", "c": "d"})

        mock_instance.post.assert_called_with(
            "https://domain/api/v2/tickets/email-verification",
            data={"a": "b", "c": "d"},
        )

    @mock.patch("auth0.management.tickets.RestClient")
    def test_pswd(self, mock_rc):
        mock_instance = mock_rc.return_value

        t = Tickets(domain="domain", token="jwttoken")
        t.create_pswd_change({"a": "b", "c": "d"})

        mock_instance.post.assert_called_with(
            "https://domain/api/v2/tickets/password-change", data={"a": "b", "c": "d"}
        )



import unittest
from unittest import mock

from ...management.users_by_email import UsersByEmail


class TestUsersByEmail(unittest.TestCase):
    def test_init_with_optionals(self):
        t = UsersByEmail(
            domain="domain", token="jwttoken", telemetry=False, timeout=(10, 2)
        )
        self.assertEqual(t.client.options.timeout, (10, 2))
        telemetry_header = t.client.base_headers.get("Auth0-Client", None)
        self.assertEqual(telemetry_header, None)

    @mock.patch("auth0.management.users_by_email.RestClient")
    def test_search_users_by_email(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = UsersByEmail(domain="domain", token="jwttoken")
        u.search_users_by_email("A@B.com")

        args, kwargs = mock_instance.get.call_args

        self.assertEqual("https://domain/api/v2/users-by-email", args[0])
        self.assertEqual(
            kwargs["params"],
            {"email": "A@B.com", "fields": None, "include_fields": "true"},
        )

        u.search_users_by_email(
            email="a@b.com", fields=["a", "b"], include_fields=False
        )

        args, kwargs = mock_instance.get.call_args

        self.assertEqual("https://domain/api/v2/users-by-email", args[0])
        self.assertEqual(
            kwargs["params"],
            {"email": "a@b.com", "fields": "a,b", "include_fields": "false"},
        )



from ..utils import is_async_available
from .actions import Actions
from .attack_protection import AttackProtection
from .blacklists import Blacklists
from .branding import Branding
from .client_credentials import ClientCredentials
from .client_grants import ClientGrants
from .clients import Clients
from .connections import Connections
from .custom_domains import CustomDomains
from .device_credentials import DeviceCredentials
from .email_templates import EmailTemplates
from .emails import Emails
from .grants import Grants
from .guardian import Guardian
from .hooks import Hooks
from .jobs import Jobs
from .log_streams import LogStreams
from .logs import Logs
from .organizations import Organizations
from .prompts import Prompts
from .resource_servers import ResourceServers
from .roles import Roles
from .rules import Rules
from .rules_configs import RulesConfigs
from .stats import Stats
from .tenants import Tenants
from .tickets import Tickets
from .user_blocks import UserBlocks
from .users import Users
from .users_by_email import UsersByEmail

modules = {
    "actions": Actions,
    "attack_protection": AttackProtection,
    "blacklists": Blacklists,
    "branding": Branding,
    "client_credentials": ClientCredentials,
    "client_grants": ClientGrants,
    "clients": Clients,
    "connections": Connections,
    "custom_domains": CustomDomains,
    "device_credentials": DeviceCredentials,
    "email_templates": EmailTemplates,
    "emails": Emails,
    "grants": Grants,
    "guardian": Guardian,
    "hooks": Hooks,
    "jobs": Jobs,
    "log_streams": LogStreams,
    "logs": Logs,
    "organizations": Organizations,
    "prompts": Prompts,
    "resource_servers": ResourceServers,
    "roles": Roles,
    "rules_configs": RulesConfigs,
    "rules": Rules,
    "stats": Stats,
    "tenants": Tenants,
    "tickets": Tickets,
    "user_blocks": UserBlocks,
    "users_by_email": UsersByEmail,
    "users": Users,
}


class Auth0:
    """Provides easy access to all endpoint classes

    Args:
        domain (str): Your Auth0 domain, e.g: 'username.auth0.com'

        token (str): Management API v2 Token

        rest_options (RestClientOptions): Pass an instance of
            RestClientOptions to configure additional RestClient
            options, such as rate-limit retries.
            (defaults to None)
    """

    def __init__(self, domain, token, rest_options=None):
        for name, cls in modules.items():
            setattr(
                self,
                name,
                cls(domain=domain, token=token, rest_options=rest_options),
            )



import unittest
from unittest import mock

from ...management.custom_domains import CustomDomains


class TestCustomDomains(unittest.TestCase):
    def test_init_with_optionals(self):
        t = CustomDomains(
            domain="domain", token="jwttoken", telemetry=False, timeout=(10, 2)
        )
        self.assertEqual(t.client.options.timeout, (10, 2))
        telemetry_header = t.client.base_headers.get("Auth0-Client", None)
        self.assertEqual(telemetry_header, None)

    @mock.patch("auth0.management.custom_domains.RestClient")
    def test_get_all(self, mock_rc):
        mock_instance = mock_rc.return_value

        g = CustomDomains(domain="domain", token="jwttoken")
        g.all()

        mock_instance.get.assert_called_with("https://domain/api/v2/custom-domains")

    @mock.patch("auth0.management.custom_domains.RestClient")
    def test_create_new(self, mock_rc):
        mock_instance = mock_rc.return_value

        g = CustomDomains(domain="domain", token="jwttoken")
        g.create_new(body={"a": "b", "c": "d", "e": "f"})

        args, kwargs = mock_instance.post.call_args

        self.assertEqual("https://domain/api/v2/custom-domains", args[0])
        self.assertEqual(kwargs["data"], {"a": "b", "c": "d", "e": "f"})

    @mock.patch("auth0.management.custom_domains.RestClient")
    def test_get_domain_by_id(self, mock_rc):
        mock_instance = mock_rc.return_value

        g = CustomDomains(domain="domain", token="jwttoken")
        g.get("an-id")

        mock_instance.get.assert_called_with(
            "https://domain/api/v2/custom-domains/an-id"
        )

    @mock.patch("auth0.management.custom_domains.RestClient")
    def test_verify(self, mock_rc):
        mock_instance = mock_rc.return_value

        g = CustomDomains(domain="domain", token="jwttoken")
        g.verify("an-id")

        args, kwargs = mock_instance.post.call_args

        self.assertEqual("https://domain/api/v2/custom-domains/an-id/verify", args[0])



import unittest
from unittest import mock

from ...management.clients import Clients


class TestClients(unittest.TestCase):
    def test_init_with_optionals(self):
        t = Clients(domain="domain", token="jwttoken", telemetry=False, timeout=(10, 2))
        self.assertEqual(t.client.options.timeout, (10, 2))
        telemetry_header = t.client.base_headers.get("Auth0-Client", None)
        self.assertEqual(telemetry_header, None)

    @mock.patch("auth0.management.clients.RestClient")
    def test_all(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = Clients(domain="domain", token="jwttoken")

        # Default parameters are requested
        c.all()

        args, kwargs = mock_instance.get.call_args

        self.assertEqual("https://domain/api/v2/clients", args[0])
        self.assertEqual(
            kwargs["params"],
            {"fields": None, "include_fields": "true", "page": None, "per_page": None},
        )

        # Fields filter
        c.all(fields=["a", "b"], include_fields=False)

        args, kwargs = mock_instance.get.call_args

        self.assertEqual("https://domain/api/v2/clients", args[0])
        self.assertEqual(
            kwargs["params"],
            {
                "fields": "a,b",
                "include_fields": "false",
                "page": None,
                "per_page": None,
            },
        )

        # Specific pagination
        c.all(page=7, per_page=25)

        args, kwargs = mock_instance.get.call_args

        self.assertEqual("https://domain/api/v2/clients", args[0])
        self.assertEqual(
            kwargs["params"],
            {"fields": None, "include_fields": "true", "page": 7, "per_page": 25},
        )

        # Extra parameters
        c.all(extra_params={"some_key": "some_value"})

        args, kwargs = mock_instance.get.call_args

        self.assertEqual("https://domain/api/v2/clients", args[0])
        self.assertEqual(
            kwargs["params"],
            {
                "fields": None,
                "include_fields": "true",
                "page": None,
                "per_page": None,
                "some_key": "some_value",
            },
        )

    @mock.patch("auth0.management.clients.RestClient")
    def test_create(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = Clients(domain="domain", token="jwttoken")
        c.create({"a": "b", "c": "d"})

        mock_instance.post.assert_called_with(
            "https://domain/api/v2/clients", data={"a": "b", "c": "d"}
        )

    @mock.patch("auth0.management.clients.RestClient")
    def test_get(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = Clients(domain="domain", token="jwttoken")
        c.get("this-id")

        args, kwargs = mock_instance.get.call_args

        self.assertEqual("https://domain/api/v2/clients/this-id", args[0])
        self.assertEqual(kwargs["params"], {"fields": None, "include_fields": "true"})

        c.get("this-id", fields=["a", "b"], include_fields=False)

        args, kwargs = mock_instance.get.call_args

        self.assertEqual("https://domain/api/v2/clients/this-id", args[0])
        self.assertEqual(kwargs["params"], {"fields": "a,b", "include_fields": "false"})

    @mock.patch("auth0.management.clients.RestClient")
    def test_delete(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = Clients(domain="domain", token="jwttoken")
        c.delete("this-id")

        mock_instance.delete.assert_called_with("https://domain/api/v2/clients/this-id")

    @mock.patch("auth0.management.clients.RestClient")
    def test_update(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = Clients(domain="domain", token="jwttoken")
        c.update("this-id", {"a": "b", "c": "d"})

        args, kwargs = mock_instance.patch.call_args

        self.assertEqual("https://domain/api/v2/clients/this-id", args[0])
        self.assertEqual(kwargs["data"], {"a": "b", "c": "d"})

    @mock.patch("auth0.management.clients.RestClient")
    def test_rotate_secret(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = Clients(domain="domain", token="jwttoken")
        c.rotate_secret("this-id")

        mock_instance.post.assert_called_with(
            "https://domain/api/v2/clients/this-id/rotate-secret",
            data={"id": "this-id"},
        )



import unittest
from unittest import mock

from ...management.stats import Stats


class TestStats(unittest.TestCase):
    def test_init_with_optionals(self):
        t = Stats(domain="domain", token="jwttoken", telemetry=False, timeout=(10, 2))
        self.assertEqual(t.client.options.timeout, (10, 2))
        telemetry_header = t.client.base_headers.get("Auth0-Client", None)
        self.assertEqual(telemetry_header, None)

    @mock.patch("auth0.management.stats.RestClient")
    def test_active_users(self, mock_rc):
        mock_instance = mock_rc.return_value

        s = Stats(domain="domain", token="jwttoken")
        s.active_users()

        mock_instance.get.assert_called_with(
            "https://domain/api/v2/stats/active-users",
        )

    @mock.patch("auth0.management.stats.RestClient")
    def test_daily_stats(self, mock_rc):
        mock_instance = mock_rc.return_value

        s = Stats(domain="domain", token="jwttoken")
        s.daily_stats()

        mock_instance.get.assert_called_with(
            "https://domain/api/v2/stats/daily",
            params={"from": None, "to": None},
        )

        s.daily_stats(from_date="12341212", to_date="56785656")

        mock_instance.get.assert_called_with(
            "https://domain/api/v2/stats/daily",
            params={"from": "12341212", "to": "56785656"},
        )



import os
import re

from setuptools import find_packages, setup


def find_version():
    file_dir = os.path.dirname(__file__)
    with open(os.path.join(file_dir, "auth0", "__init__.py")) as f:
        version = re.search(r'^__version__ = [\'"]([^\'"]*)[\'"]', f.read())
        if version:
            return version.group(1)
        else:
            raise RuntimeError("Unable to find version string.")


with open("README.md", encoding="utf-8") as f:
    long_description = f.read()


setup(
    name="auth0-python",
    version=find_version(),
    description="Auth0 Python SDK",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Auth0",
    author_email="support@auth0.com",
    license="MIT",
    packages=find_packages(),
    install_requires=["requests>=2.14.0", "pyjwt[crypto]>=2.6.0"],
    extras_require={"test": ["coverage", "pre-commit"]},
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    url="https://github.com/auth0/auth0-python",
)



from ..rest import RestClient


class ResourceServers:
    """Auth0 resource servers endpoints

    Args:
        domain (str): Your Auth0 domain, e.g: 'username.auth0.com'

        token (str): Management API v2 Token

        telemetry (bool, optional): Enable or disable Telemetry
            (defaults to True)

        timeout (float or tuple, optional): Change the requests
            connect and read timeout. Pass a tuple to specify
            both values separately or a float to set both to it.
            (defaults to 5.0 for both)

        rest_options (RestClientOptions): Pass an instance of
            RestClientOptions to configure additional RestClient
            options, such as rate-limit retries.
            (defaults to None)
    """

    def __init__(
        self,
        domain,
        token,
        telemetry=True,
        timeout=5.0,
        protocol="https",
        rest_options=None,
    ):
        self.domain = domain
        self.protocol = protocol
        self.client = RestClient(
            jwt=token, telemetry=telemetry, timeout=timeout, options=rest_options
        )

    def _url(self, id=None):
        url = f"{self.protocol}://{self.domain}/api/v2/resource-servers"
        if id is not None:
            return f"{url}/{id}"
        return url

    def create(self, body):
        """Create a new resource server.

        Args:
           body (dict): Attributes for the new resource Server.

        See: https://auth0.com/docs/api/management/v2#!/Resource_Servers/post_resource_servers
        """

        return self.client.post(self._url(), data=body)

    def get_all(self, page=None, per_page=None, include_totals=False):
        """Retrieves all resource servers

        Args:
            page (int, optional): The result's page number (zero based). When not set,
              the default value is up to the server.

            per_page (int, optional): The amount of entries per page. When not set,
              the default value is up to the server.

            include_totals (bool, optional): True if the query summary is
                to be included in the result, False otherwise. Defaults to False.


        See: https://auth0.com/docs/api/management/v2#!/Resource_Servers/get_resource_servers
        """

        params = {
            "page": page,
            "per_page": per_page,
            "include_totals": str(include_totals).lower(),
        }

        return self.client.get(self._url(), params=params)

    def get(self, id):
        """Retrieves a resource server by its id.

        Args:
           id (str): id of the resource server to get.


        See: https://auth0.com/docs/api/management/v2#!/Resource_Servers/get_resource_servers_by_id
        """

        return self.client.get(self._url(id))

    def delete(self, id):
        """Deletes a resource server.

        Args:
           id (str): Id of resource server to delete.


        See: https://auth0.com/docs/api/management/v2#!/Resource_Servers/delete_resource_servers_by_id
        """

        return self.client.delete(self._url(id))

    def update(self, id, body):
        """Modifies a resource server.

        Args:
           id (str): The id of the resource server to update.

           body (dict): Attributes to modify.

        See: https://auth0.com/docs/api/management/v2#!/Resource_Servers/patch_resource_servers_by_id
        """

        return self.client.patch(self._url(id), data=body)



from ..rest import RestClient


class AttackProtection:
    """Auth0 attack protection endpoints

    Args:
        domain (str): Your Auth0 domain, e.g: 'username.auth0.com'

        token (str): Management API v2 Token

        telemetry (bool, optional): Enable or disable Telemetry
            (defaults to True)

        timeout (float or tuple, optional): Change the requests
            connect and read timeout. Pass a tuple to specify
            both values separately or a float to set both to it.
            (defaults to 5.0 for both)

        rest_options (RestClientOptions): Pass an instance of
            RestClientOptions to configure additional RestClient
            options, such as rate-limit retries.
            (defaults to None)
    """

    def __init__(
        self,
        domain,
        token,
        telemetry=True,
        timeout=5.0,
        protocol="https",
        rest_options=None,
    ):
        self.domain = domain
        self.protocol = protocol
        self.client = RestClient(
            jwt=token, telemetry=telemetry, timeout=timeout, options=rest_options
        )

    def _url(self, component):
        return "{}://{}/api/v2/attack-protection/{}".format(
            self.protocol, self.domain, component
        )

    def get_breached_password_detection(self):
        """Get breached password detection settings.

        Returns the breached password detection settings.

        See: https://auth0.com/docs/api/management/v2#!/Attack_Protection/get_breached_password_detection
        """
        url = self._url("breached-password-detection")
        return self.client.get(url)

    def update_breached_password_detection(self, body):
        """Update breached password detection settings.

        Returns the breached password detection settings.

        Args:

           body (dict): breached password detection settings.

        See: https://auth0.com/docs/api/management/v2#!/Attack_Protection/patch_breached_password_detection
        """
        url = self._url("breached-password-detection")
        return self.client.patch(url, data=body)

    def get_brute_force_protection(self):
        """Get the brute force configuration.

        Returns the brute force configuration.

        See: https://auth0.com/docs/api/management/v2#!/Attack_Protection/get_brute_force_protection
        """
        url = self._url("brute-force-protection")
        return self.client.get(url)

    def update_brute_force_protection(self, body):
        """Update the brute force configuration.

        Returns the brute force configuration.

        Args:

           body (dict): updates of the brute force configuration.

        See: https://auth0.com/docs/api/management/v2#!/Attack_Protection/patch_brute_force_protection
        """
        url = self._url("brute-force-protection")
        return self.client.patch(url, data=body)

    def get_suspicious_ip_throttling(self):
        """Get the suspicious IP throttling configuration.

        Returns the suspicious IP throttling configuration.

        See: https://auth0.com/docs/api/management/v2#!/Attack_Protection/get_suspicious_ip_throttling
        """
        url = self._url("suspicious-ip-throttling")
        return self.client.get(url)

    def update_suspicious_ip_throttling(self, body):
        """Update the suspicious IP throttling configuration.

        Returns the suspicious IP throttling configuration.

        Args:

           body (dict): updates of the suspicious IP throttling configuration.

        See: https://auth0.com/docs/api/management/v2#!/Attack_Protection/patch_suspicious_ip_throttling
        """
        url = self._url("suspicious-ip-throttling")
        return self.client.patch(url, data=body)



import base64
import json
import platform
import re
import sys
import unittest
from tempfile import TemporaryFile
from unittest.mock import ANY, MagicMock

import aiohttp
from aioresponses import CallbackResult, aioresponses
from callee import Attrs

from auth0.asyncify import asyncify
from auth0.management import Clients, Guardian, Jobs

clients = re.compile(r"^https://example\.com/api/v2/clients.*")
factors = re.compile(r"^https://example\.com/api/v2/guardian/factors.*")
users_imports = re.compile(r"^https://example\.com/api/v2/jobs/users-imports.*")
payload = {"foo": "bar"}

telemetry = base64.b64encode(
    json.dumps(
        {
            "name": "auth0-python",
            "version": sys.modules["auth0"].__version__,
            "env": {
                "python": platform.python_version(),
            },
        }
    ).encode("utf-8")
).decode()

headers = {
    "User-Agent": f"Python/{platform.python_version()}",
    "Authorization": "Bearer jwt",
    "Content-Type": "application/json",
    "Auth0-Client": telemetry,
}


def get_callback(status=200, response=None):
    mock = MagicMock(
        return_value=CallbackResult(status=status, payload=response or payload)
    )

    def callback(url, **kwargs):
        return mock(url, **kwargs)

    return callback, mock


@unittest.skipIf(
    not hasattr(unittest, "IsolatedAsyncioTestCase"),
    "python 3.7 doesn't have IsolatedAsyncioTestCase",
)
class TestAsyncify(getattr(unittest, "IsolatedAsyncioTestCase", object)):
    @aioresponses()
    async def test_get(self, mocked):
        callback, mock = get_callback()
        mocked.get(clients, callback=callback)
        c = asyncify(Clients)(domain="example.com", token="jwt")
        self.assertEqual(await c.all_async(), payload)
        mock.assert_called_with(
            Attrs(path="/api/v2/clients"),
            allow_redirects=True,
            params={"include_fields": "true"},
            headers=headers,
            timeout=ANY,
        )

    @aioresponses()
    async def test_post(self, mocked):
        callback, mock = get_callback()
        mocked.post(clients, callback=callback)
        c = asyncify(Clients)(domain="example.com", token="jwt")
        data = {"client": 1}
        self.assertEqual(await c.create_async(data), payload)
        mock.assert_called_with(
            Attrs(path="/api/v2/clients"),
            allow_redirects=True,
            json=data,
            headers=headers,
            timeout=ANY,
        )

    @aioresponses()
    async def test_file_post(self, mocked):
        callback, mock = get_callback()
        mocked.post(users_imports, callback=callback)
        j = asyncify(Jobs)(domain="example.com", token="jwt")
        users = TemporaryFile()
        self.assertEqual(await j.import_users_async("connection-1", users), payload)
        file_port_headers = headers.copy()
        file_port_headers.pop("Content-Type")
        mock.assert_called_with(
            Attrs(path="/api/v2/jobs/users-imports"),
            allow_redirects=True,
            data={
                "connection_id": "connection-1",
                "upsert": "false",
                "send_completion_email": "true",
                "external_id": None,
                "users": users,
            },
            headers=file_port_headers,
            timeout=ANY,
        )
        users.close()

    @aioresponses()
    async def test_patch(self, mocked):
        callback, mock = get_callback()
        mocked.patch(clients, callback=callback)
        c = asyncify(Clients)(domain="example.com", token="jwt")
        data = {"client": 1}
        self.assertEqual(await c.update_async("client-1", data), payload)
        mock.assert_called_with(
            Attrs(path="/api/v2/clients/client-1"),
            allow_redirects=True,
            json=data,
            headers=headers,
            timeout=ANY,
        )

    @aioresponses()
    async def test_put(self, mocked):
        callback, mock = get_callback()
        mocked.put(factors, callback=callback)
        g = asyncify(Guardian)(domain="example.com", token="jwt")
        data = {"factor": 1}
        self.assertEqual(await g.update_factor_async("factor-1", data), payload)
        mock.assert_called_with(
            Attrs(path="/api/v2/guardian/factors/factor-1"),
            allow_redirects=True,
            json=data,
            headers=headers,
            timeout=ANY,
        )

    @aioresponses()
    async def test_delete(self, mocked):
        callback, mock = get_callback()
        mocked.delete(clients, callback=callback)
        c = asyncify(Clients)(domain="example.com", token="jwt")
        self.assertEqual(await c.delete_async("client-1"), payload)
        mock.assert_called_with(
            Attrs(path="/api/v2/clients/client-1"),
            allow_redirects=True,
            params={},
            json=None,
            headers=headers,
            timeout=ANY,
        )

    @aioresponses()
    async def test_shared_session(self, mocked):
        callback, mock = get_callback()
        mocked.get(clients, callback=callback)
        async with asyncify(Clients)(domain="example.com", token="jwt") as c:
            self.assertEqual(await c.all_async(), payload)
        mock.assert_called_with(
            Attrs(path="/api/v2/clients"),
            allow_redirects=True,
            params={"include_fields": "true"},
            headers=headers,
            timeout=ANY,
        )

    @aioresponses()
    async def test_rate_limit(self, mocked):
        callback, mock = get_callback(status=429)
        mocked.get(clients, callback=callback)
        mocked.get(clients, callback=callback)
        mocked.get(clients, callback=callback)
        mocked.get(clients, payload=payload)
        c = asyncify(Clients)(domain="example.com", token="jwt")
        rest_client = c._async_client.client
        rest_client._skip_sleep = True
        self.assertEqual(await c.all_async(), payload)
        self.assertEqual(3, mock.call_count)
        (a, b, c) = rest_client._metrics["backoff"]
        self.assertTrue(100 <= a < b < c <= 1000)

    @aioresponses()
    async def test_timeout(self, mocked):
        callback, mock = get_callback()
        mocked.get(clients, callback=callback)
        c = asyncify(Clients)(domain="example.com", token="jwt", timeout=(8.8, 9.9))
        self.assertEqual(await c.all_async(), payload)
        mock.assert_called_with(
            ANY,
            allow_redirects=ANY,
            params=ANY,
            headers=ANY,
            timeout=aiohttp.ClientTimeout(sock_connect=8.8, sock_read=9.9),
        )



import unittest
from unittest import mock

from ...management.logs import Logs


class TestLogs(unittest.TestCase):
    def test_init_with_optionals(self):
        t = Logs(domain="domain", token="jwttoken", telemetry=False, timeout=(10, 2))
        self.assertEqual(t.client.options.timeout, (10, 2))
        telemetry_header = t.client.base_headers.get("Auth0-Client", None)
        self.assertEqual(telemetry_header, None)

    @mock.patch("auth0.management.logs.RestClient")
    def test_search(self, mock_rc):
        mock_instance = mock_rc.return_value

        logs = Logs(domain="domain", token="jwttoken")
        logs.search()

        args, kwargs = mock_instance.get.call_args
        self.assertEqual("https://domain/api/v2/logs", args[0])
        self.assertIsNone(kwargs["params"]["sort"])
        self.assertIsNone(kwargs["params"]["q"])
        self.assertIsNone(kwargs["params"]["from"])
        self.assertIsNone(kwargs["params"]["take"])
        self.assertEqual(kwargs["params"]["include_fields"], "true")
        self.assertEqual(kwargs["params"]["include_totals"], "true")
        self.assertEqual(kwargs["params"]["per_page"], 50)
        self.assertEqual(kwargs["params"]["page"], 0)
        self.assertIsNone(kwargs["params"]["fields"])

        logs.search(fields=["description", "client_id"])

        args, kwargs = mock_instance.get.call_args
        self.assertEqual(kwargs["params"]["fields"], "description,client_id")

        logs.search(page=0, per_page=2)

        args, kwargs = mock_instance.get.call_args
        self.assertEqual(kwargs["params"]["per_page"], 2)
        self.assertEqual(kwargs["params"]["page"], 0)

    @mock.patch("auth0.management.logs.RestClient")
    def test_get(self, mock_rc):
        mock_instance = mock_rc.return_value

        logs = Logs(domain="domain", token="jwttoken")
        logs.get("get_id")

        mock_instance.get.assert_called_with("https://domain/api/v2/logs/get_id")



import unittest
from unittest import mock

from ...management.client_credentials import ClientCredentials


class TestClientCredentials(unittest.TestCase):
    def test_init_with_optionals(self):
        t = ClientCredentials(
            domain="domain", token="jwttoken", telemetry=False, timeout=(10, 2)
        )
        self.assertEqual(t.client.options.timeout, (10, 2))
        telemetry_header = t.client.base_headers.get("Auth0-Client", None)
        self.assertEqual(telemetry_header, None)

    @mock.patch("auth0.management.client_credentials.RestClient")
    def test_all(self, mock_rc):
        mock_instance = mock_rc.return_value
        c = ClientCredentials(domain="domain", token="jwttoken")
        c.all("cid")
        mock_instance.get.assert_called_with(
            "https://domain/api/v2/clients/cid/credentials"
        )

    @mock.patch("auth0.management.client_credentials.RestClient")
    def test_get(self, mock_rc):
        mock_instance = mock_rc.return_value
        c = ClientCredentials(domain="domain", token="jwttoken")
        c.get("cid", "this-id")
        mock_instance.get.assert_called_with(
            "https://domain/api/v2/clients/cid/credentials/this-id"
        )

    @mock.patch("auth0.management.client_credentials.RestClient")
    def test_create(self, mock_rc):
        mock_instance = mock_rc.return_value
        c = ClientCredentials(domain="domain", token="jwttoken")
        c.create("cid", {"a": "b", "c": "d"})
        mock_instance.post.assert_called_with(
            "https://domain/api/v2/clients/cid/credentials", data={"a": "b", "c": "d"}
        )

    @mock.patch("auth0.management.client_credentials.RestClient")
    def test_delete(self, mock_rc):
        mock_instance = mock_rc.return_value
        c = ClientCredentials(domain="domain", token="jwttoken")
        c.delete("cid", "this-id")
        mock_instance.delete.assert_called_with(
            "https://domain/api/v2/clients/cid/credentials/this-id"
        )



from ..rest import RestClient


class Roles:
    """Auth0 roles endpoints

    Args:
        domain (str): Your Auth0 domain, e.g: 'username.auth0.com'

        token (str): Management API v2 Token

        telemetry (bool, optional): Enable or disable Telemetry
            (defaults to True)

        timeout (float or tuple, optional): Change the requests
            connect and read timeout. Pass a tuple to specify
            both values separately or a float to set both to it.
            (defaults to 5.0 for both)

        rest_options (RestClientOptions): Pass an instance of
            RestClientOptions to configure additional RestClient
            options, such as rate-limit retries.
            (defaults to None)
    """

    def __init__(
        self,
        domain,
        token,
        telemetry=True,
        timeout=5.0,
        protocol="https",
        rest_options=None,
    ):
        self.domain = domain
        self.protocol = protocol
        self.client = RestClient(
            jwt=token, telemetry=telemetry, timeout=timeout, options=rest_options
        )

    def _url(self, id=None):
        url = f"{self.protocol}://{self.domain}/api/v2/roles"
        if id is not None:
            return f"{url}/{id}"
        return url

    def list(self, page=0, per_page=25, include_totals=True, name_filter=None):
        """List or search roles.

        Args:
            page (int, optional): The result's page number (zero based). By default,
               retrieves the first page of results.

            per_page (int, optional): The amount of entries per page. By default,
               retrieves 25 results per page.

            include_totals (bool, optional): True if the query summary is
                to be included in the result, False otherwise. Defaults to True.

            name_filter (str, optional): A case-insensitive filter to apply
                to search for roles by name.

        See: https://auth0.com/docs/api/management/v2#!/Roles/get_roles
        """
        params = {
            "per_page": per_page,
            "page": page,
            "include_totals": str(include_totals).lower(),
            "name_filter": name_filter,
        }
        return self.client.get(self._url(), params=params)

    def create(self, body):
        """Creates a new role.

        Args:
            body (dict): the attributes for the role to create.

        See: https://auth0.com/docs/api/v2#!/Roles/post_roles
        """
        return self.client.post(self._url(), data=body)

    def get(self, id):
        """Get a role.

        Args:
            id (str): The id of the role to retrieve.

        See: https://auth0.com/docs/api/management/v2#!/Roles/get_roles_by_id
        """

        return self.client.get(self._url(id))

    def delete(self, id):
        """Delete a role.

        Args:
            id (str): The id of the role to delete.

        See: https://auth0.com/docs/api/management/v2#!/Roles/delete_roles_by_id
        """
        return self.client.delete(self._url(id))

    def update(self, id, body):
        """Update a role with the attributes passed in 'body'

        Args:
            id (str): The id of the role to update.

            body (dict): the attributes to update on the role.

        See: https://auth0.com/docs/api/management/v2#!/Roles/patch_roles_by_id
        """
        return self.client.patch(self._url(id), data=body)

    def list_users(
        self, id, page=0, per_page=25, include_totals=True, from_param=None, take=None
    ):
        """List the users that have been associated with a given role.

        Args:
            id (str): The role's id.

            page (int, optional): The result's page number (zero based). By default,
               retrieves the first page of results.

            per_page (int, optional): The amount of entries per page. By default,
               retrieves 25 results per page.

            include_totals (bool, optional): True if the query summary is
                to be included in the result, False otherwise. Defaults to True.

            from_param (str, optional): Checkpoint Id from which to begin retrieving results.
                You can limit the number of entries using the take parameter.

            take (int, optional): The total amount of entries to retrieve when
                using the from parameter. When not set, the default value is up to the server.

        See https://auth0.com/docs/api/management/v2#!/Roles/get_role_user
        """

        params = {
            "per_page": per_page,
            "page": page,
            "include_totals": str(include_totals).lower(),
            "from": from_param,
            "take": take,
        }

        url = self._url(f"{id}/users")
        return self.client.get(url, params=params)

    def add_users(self, id, users):
        """Assign users to a role.

        Args:
            id (str): The role's id.

            users (list of str): A list of users ids to add to this role.

        See https://auth0.com/docs/api/management/v2#!/Roles/post_role_users
        """
        url = self._url(f"{id}/users")
        body = {"users": users}
        return self.client.post(url, data=body)

    def list_permissions(self, id, page=0, per_page=25, include_totals=True):
        """List the permissions associated to a role.

        Args:
            id (str): The role's id.

            page (int, optional): The result's page number (zero based). By default,
               retrieves the first page of results.

            per_page (int, optional): The amount of entries per page. By default,
               retrieves 25 results per page.

            include_totals (bool, optional): True if the query summary is
                to be included in the result, False otherwise. Defaults to True.

        See https://auth0.com/docs/api/management/v2#!/Roles/get_role_permission
        """
        params = {
            "per_page": per_page,
            "page": page,
            "include_totals": str(include_totals).lower(),
        }
        url = self._url(f"{id}/permissions")
        return self.client.get(url, params=params)

    def remove_permissions(self, id, permissions):
        """Unassociates permissions from a role.

        Args:
            id (str): The role's id.

            permissions (list of str): A list of permission ids to unassociate from the role.

        See https://auth0.com/docs/api/management/v2#!/Roles/delete_role_permission_assignment
        """
        url = self._url(f"{id}/permissions")
        body = {"permissions": permissions}
        return self.client.delete(url, data=body)

    def add_permissions(self, id, permissions):
        """Associates permissions with a role.

        Args:
            id (str): The role's id.

            permissions (list of str): A list of permission ids to associate to the role.

        See https://auth0.com/docs/api/management/v2#!/Roles/post_role_permission_assignment
        """
        url = self._url(f"{id}/permissions")
        body = {"permissions": permissions}
        return self.client.post(url, data=body)



import re
import unittest
from unittest.mock import ANY, MagicMock

from aioresponses import CallbackResult, aioresponses
from callee import Attrs

from auth0.management.async_auth0 import AsyncAuth0 as Auth0

clients = re.compile(r"^https://example\.com/api/v2/clients.*")
factors = re.compile(r"^https://example\.com/api/v2/guardian/factors.*")
payload = {"foo": "bar"}


def get_callback(status=200):
    mock = MagicMock(return_value=CallbackResult(status=status, payload=payload))

    def callback(url, **kwargs):
        return mock(url, **kwargs)

    return callback, mock


@unittest.skipIf(
    not hasattr(unittest, "IsolatedAsyncioTestCase"),
    "python 3.7 doesn't have IsolatedAsyncioTestCase",
)
class TestAuth0(getattr(unittest, "IsolatedAsyncioTestCase", object)):
    @aioresponses()
    async def test_get(self, mocked):
        callback, mock = get_callback()
        mocked.get(clients, callback=callback)
        auth0 = Auth0(domain="example.com", token="jwt")
        self.assertEqual(await auth0.clients.all_async(), payload)
        mock.assert_called_with(
            Attrs(path="/api/v2/clients"),
            allow_redirects=True,
            params={"include_fields": "true"},
            headers=ANY,
            timeout=ANY,
        )

    @aioresponses()
    async def test_shared_session(self, mocked):
        callback, mock = get_callback()
        callback2, mock2 = get_callback()
        mocked.get(clients, callback=callback)
        mocked.put(factors, callback=callback2)
        async with Auth0(domain="example.com", token="jwt") as auth0:
            self.assertEqual(await auth0.clients.all_async(), payload)
            self.assertEqual(
                await auth0.guardian.update_factor_async("factor-1", {"factor": 1}),
                payload,
            )
        mock.assert_called_with(
            Attrs(path="/api/v2/clients"),
            allow_redirects=True,
            params={"include_fields": "true"},
            headers=ANY,
            timeout=ANY,
        )
        mock2.assert_called_with(
            Attrs(path="/api/v2/guardian/factors/factor-1"),
            allow_redirects=True,
            json={"factor": 1},
            headers=ANY,
            timeout=ANY,
        )



import unittest
from unittest import mock

from ...management.rules import Rules


class TestRules(unittest.TestCase):
    def test_init_with_optionals(self):
        t = Rules(domain="domain", token="jwttoken", telemetry=False, timeout=(10, 2))
        self.assertEqual(t.client.options.timeout, (10, 2))
        telemetry_header = t.client.base_headers.get("Auth0-Client", None)
        self.assertEqual(telemetry_header, None)

    @mock.patch("auth0.management.rules.RestClient")
    def test_all(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = Rules(domain="domain", token="jwttoken")

        # with default params
        c.all()

        args, kwargs = mock_instance.get.call_args

        self.assertEqual("https://domain/api/v2/rules", args[0])
        self.assertEqual(
            kwargs["params"],
            {
                "fields": None,
                "include_fields": "true",
                "enabled": "true",
                "stage": "login_success",
                "page": None,
                "per_page": None,
                "include_totals": "false",
            },
        )

        # with stage and fields params
        c.all(stage="stage", enabled=False, fields=["a", "b"], include_fields=False)

        args, kwargs = mock_instance.get.call_args

        self.assertEqual("https://domain/api/v2/rules", args[0])
        self.assertEqual(
            kwargs["params"],
            {
                "fields": "a,b",
                "include_fields": "false",
                "enabled": "false",
                "stage": "stage",
                "page": None,
                "per_page": None,
                "include_totals": "false",
            },
        )

        # with pagination params
        c.all(page=3, per_page=27, include_totals=True)

        args, kwargs = mock_instance.get.call_args

        self.assertEqual("https://domain/api/v2/rules", args[0])
        self.assertEqual(
            kwargs["params"],
            {
                "fields": None,
                "include_fields": "true",
                "enabled": "true",
                "stage": "login_success",
                "page": 3,
                "per_page": 27,
                "include_totals": "true",
            },
        )

    @mock.patch("auth0.management.rules.RestClient")
    def test_create(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = Rules(domain="domain", token="jwttoken")
        c.create({"a": "b", "c": "d"})

        args, kwargs = mock_instance.post.call_args

        self.assertEqual("https://domain/api/v2/rules", args[0])
        self.assertEqual(kwargs["data"], {"a": "b", "c": "d"})

    @mock.patch("auth0.management.rules.RestClient")
    def test_get(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = Rules(domain="domain", token="jwttoken")
        c.get("an-id")

        args, kwargs = mock_instance.get.call_args

        self.assertEqual("https://domain/api/v2/rules/an-id", args[0])
        self.assertEqual(kwargs["params"], {"fields": None, "include_fields": "true"})

        c.get("an-id", fields=["a", "b"], include_fields=False)

        args, kwargs = mock_instance.get.call_args

        self.assertEqual("https://domain/api/v2/rules/an-id", args[0])
        self.assertEqual(kwargs["params"], {"fields": "a,b", "include_fields": "false"})

    @mock.patch("auth0.management.rules.RestClient")
    def test_delete(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = Rules(domain="domain", token="jwttoken")
        c.delete("an-id")

        mock_instance.delete.assert_called_with("https://domain/api/v2/rules/an-id")

    @mock.patch("auth0.management.rules.RestClient")
    def test_update(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = Rules(domain="domain", token="jwttoken")
        c.update("an-id", {"a": "b", "c": "d"})

        args, kwargs = mock_instance.patch.call_args

        self.assertEqual("https://domain/api/v2/rules/an-id", args[0])
        self.assertEqual(kwargs["data"], {"a": "b", "c": "d"})
